@@ -1,9 +1,9 @@
 #!/usr/bin/wish -f
 #
-# $Id: TdDebug.tcl,v 4.1 1994/01/30 21:03:30 schmid Exp schmid $
+# $Id: TdDebug.tcl,v 5.0 1995/05/02 06:32:04 gs Exp $
 #
 # tdebug.tcl - A simple debugger for tcl scripts
-# Version 0.3
+# Version 1.0
 #
 # Copyright (C) 1993 Gregor Schmid 
 #
@@ -66,12 +66,12 @@ foreach i "\
 				\"blt_busy\"] \
 					0} \
 	{preparedtag 		{-background [lindex [.td_dummy configure\
-				-activebackground] 4]}} \
+				-activebackground] 4]} NOCHECK} \
 	{activetag              {-background orange -relief raised \
-				-borderwidth 1}} \
+				-borderwidth 1} NOCHECK} \
 	{breaktag		{-foreground red -background gold -relief raised \
-				-borderwidth 1}} \
-	{tagpriority		{prepared sel active break}} \
+				-borderwidth 1} NOCHECK} \
+	{tagpriority		{prepared sel active break} NOCHECK} \
 " {
     if {[lindex $i 2] != "NOCHECK"} {
 	if {![info exists td_priv([lindex $i 0])] || \
@@ -276,7 +276,7 @@ proc td_parseScript {script name {start 0} {end -1}} {
 		set line [td_parseLine $line $name $lnum]
 	    }
 	    set rline "td_eval $name $lnum.0 $lnum.end {} ; $line"
-	} elseif { [info complete $line] } {
+	} elseif { [info complete $line] && $lnum > $switch } {
 	    if $high_detail {
 		set line [td_parseLine $line $name $lnum]
 	    }
@@ -302,13 +302,18 @@ proc td_parseScript {script name {start 0} {end -1}} {
 		incr l
 	    }
 	    set switch $l
+	    puts $switch
 	} elseif {$lnum > $switch} {
 	    if $high_detail {
 		set line [td_parseLine $line $name $lnum] 
 	    }
 	    set rline "td_eval $name $lnum.0 $lnum.end {} ; $line"
 	} else {
+	    if $high_detail {
+		set line [td_parseLine $line $name $lnum]
+	    }
 	    set rline $line
+	    puts $rline
 	}
 	append result "$rline\n"
     }
@@ -504,6 +509,7 @@ proc td_eval {name l1 l2 script} {
     # pop up debugger window
     if {[wm state $td_Top] != "normal"} {
 	wm deiconify $td_Top
+	raise $td_Top
     }
     td_displayProc $name
     $td_Listing tag remove active 1.0 end
@@ -1002,13 +1008,16 @@ proc td_evalLine {} {
 
 proc td_preparedProcs {} {
     set names [info procs *]
-    set procs ""
+    set normal ""
+    set prepared ""
     foreach i $names {
 	if {[string match #tdebug* [info body $i]]} {
-	    lappend procs $i
+	    lappend prepared $i
+	} else {
+	    lappend normal $i
 	}
     }
-    return [lsort $procs]
+    return [list [lsort $normal] [lsort $prepared]]
 }
 
 # }}}
@@ -1211,6 +1220,9 @@ proc td_displayProc {name} {
 proc td_detachDebugger {} {
     global td_Top td_BT td_WH td_ET td_Choose
 
+    foreach i [lindex [td_preparedProcs] 1] {
+	td_restoreProc $i
+    }
     rename proc ""
     rename td_origProc proc
     foreach i "$td_Top $td_BT $td_WH $td_ET" {
@@ -1247,7 +1259,7 @@ proc td_zzz {} {}
 #	size		number of lines to keep for history (default 20)
 
 proc td_cline_prepare {w {size 20}} {
-    global td_cline_priv
+    global td_cline_priv tk_version
 
     set td_cline_priv(Total$w) $size
     set td_cline_priv(Next$w) 0
@@ -1260,27 +1272,44 @@ proc td_cline_prepare {w {size 20}} {
     bind $w <Up> {td_cline_previous %W}
     bind $w <Control-n> {td_cline_next %W}
     bind $w <Down> {td_cline_next %W}
-    bind $w <Enter> {focus %W}
-    bind $w <Leave> {focus [focus default]}
-    bind $w <Control-b> {
-	%W icursor [expr [%W index insert] - 1]
-	tk_entrySeeCaret %W
+    if {$tk_version < 4.0} {
+	bind $w <Enter> {focus %W}
+	bind $w <Leave> {focus [focus default]}
+	bind $w <Control-b> {
+	    %W icursor [expr [%W index insert] - 1]
+	    tk_entrySeeCaret %W
+	}
+	bind $w <Left> {
+	    %W icursor [expr [%W index insert] - 1]
+	    tk_entrySeeCaret %W
+	}
+	bind $w <Control-f> {
+	    %W icursor [expr [%W index insert] + 1] ;
+	    tk_entrySeeCaret %W
+	}
+	bind $w <Right> {
+	    %W icursor [expr [%W index insert] + 1]
+	    tk_entrySeeCaret %W
+	}
+	bind $w <Control-a> {%W icursor 0 ; tk_entrySeeCaret %W}
+	bind $w <Home> {%W icursor 0 ; tk_entrySeeCaret %W}
+	bind $w <Control-e> {%W icursor end ; tk_entrySeeCaret %W}
+	bind $w <End> {%W icursor end ; tk_entrySeeCaret %W}
+	bind $w <Control-k> {td_cline_kill %W}
+	bind $w <Control-y> {
+	    %W insert insert $td_cline_priv(Kill%W)
+	    tk_entrySeeCaret %W
+	}
+	bind $w <Control-d> {%W delete [%W index insert] [%W index insert]}
+	bind $w <Delete> {%W delete [%W index insert] [%W index insert]}
+    } else {
+	bind $w <Control-k> {td_cline_kill %W}
+	bind $w <Control-y> {
+	    %W insert insert $td_cline_priv(Kill%W)
+	    tkEntrySeeInsert %W
+	}
+	bindtags $w "$w Entry all"
     }
-    bind $w <Left> {%W icursor [expr [%W index insert] - 1] ; tk_entrySeeCaret %W}
-    bind $w <Control-f> {
-	%W icursor [expr [%W index insert] + 1] ;
-	tk_entrySeeCaret %W
-    }
-    bind $w <Right> {%W icursor [expr [%W index insert] + 1] ; tk_entrySeeCaret %W}
-    bind $w <Control-a> {%W icursor 0 ; tk_entrySeeCaret %W}
-    bind $w <Home> {%W icursor 0 ; tk_entrySeeCaret %W}
-    bind $w <Control-e> {%W icursor end ; tk_entrySeeCaret %W}
-    bind $w <End> {%W icursor end ; tk_entrySeeCaret %W}
-    bind $w <Control-k> {td_cline_kill %W}
-    bind $w <Control-y> {%W insert insert $td_cline_priv(Kill%W) ; tk_entrySeeCaret %W}
-    bind $w <Control-d> {%W delete [%W index insert] [%W index insert]}
-    bind $w <Delete> {%W delete [%W index insert] [%W index insert]}
-
 }
 
 # }}}
@@ -1313,7 +1342,7 @@ proc td_cline_return {w} {
 #	w		name of entry widget
 
 proc td_cline_previous {w} {
-    global td_cline_priv
+    global td_cline_priv tk_version
 
     if {$td_cline_priv(Current$w) == $td_cline_priv(Next$w)} {
 	set td_cline_priv($td_cline_priv(Next$w)$w) [$w get]
@@ -1329,7 +1358,11 @@ proc td_cline_previous {w} {
     } 
     $w delete 0 end
     $w insert 0 $td_cline_priv($td_cline_priv(Current$w)$w)
-    tk_entrySeeCaret $w
+    if {$tk_version >= 4.0} {
+	tkEntrySeeInsert $w
+    } else {
+	tk_entrySeeCaret $w
+    }
 }
 
 # }}}
@@ -1340,7 +1373,7 @@ proc td_cline_previous {w} {
 #	w		name of entry widget
 
 proc td_cline_next {w} {
-    global td_cline_priv
+    global td_cline_priv tk_version
 
     if {$td_cline_priv(Current$w) == $td_cline_priv(Next$w)} {return}
     incr td_cline_priv(Current$w) 
@@ -1349,8 +1382,11 @@ proc td_cline_next {w} {
     } 
     $w delete 0 end
     $w insert 0 $td_cline_priv($td_cline_priv(Current$w)$w)
-    tk_entrySeeCaret $w
-}
+    if {$tk_version >= 4.0} {
+	tkEntrySeeInsert $w
+    } else {
+	tk_entrySeeCaret $w
+}   }
 
 # }}}
 # {{{ td_cline_kill
@@ -1401,6 +1437,9 @@ set td_ListProc		$td_ListNameFrame.listProc
 set td_TextFrame	$td_ListFrame.textFrame
 set td_Listing		$td_TextFrame.listing
 set td_ListScroll	$td_TextFrame.listScroll
+set td_ListHSFrame	$td_ListFrame.hscrollFrame
+set td_ListHScroll	$td_ListHSFrame.scroll
+set td_ListHFill	$td_ListHSFrame.fill
 
 set td_VarFrame		$td_MainRegion.varFrame
 set td_VarFrame1	$td_VarFrame.varFrame1
@@ -1437,20 +1476,32 @@ set td_DelayMore	$td_DelayFrame.delayMore
 
 set td_BT		.td_BT
 set td_BTMain		$td_BT.main
-set td_BTText		$td_BTMain.text
-set td_BTScroll		$td_BTMain.scroll
+set td_BTTop		$td_BTMain.top
+set td_BTText		$td_BTTop.text
+set td_BTScroll		$td_BTTop.scroll
+set td_BTSHFrame	$td_BTMain.hScroll
+set td_BTScrollH	$td_BTSHFrame.scroll
+set td_BTFillH		$td_BTSHFrame.fill
 set td_BTClose		$td_BT.close
 
 set td_WH		.td_WH
 set td_WHMain		$td_WH.main
-set td_WHText		$td_WHMain.text
-set td_WHScroll		$td_WHMain.scroll
+set td_WHTop		$td_WHMain.top
+set td_WHText		$td_WHTop.text
+set td_WHScroll		$td_WHTop.scroll
+set td_WHSHFrame	$td_WHMain.hScroll
+set td_WHScrollH	$td_WHSHFrame.scroll
+set td_WHFillH		$td_WHSHFrame.fill
 set td_WHClose		$td_WH.close
 
 set td_ET		.td_ET
 set td_ETMain		$td_ET.main
-set td_ETText		$td_ETMain.text
-set td_ETScroll		$td_ETMain.scroll
+set td_ETTop		$td_ETMain.top
+set td_ETText		$td_ETTop.text
+set td_ETScroll		$td_ETTop.scroll
+set td_ETSHFrame	$td_ETMain.hScroll
+set td_ETScrollH	$td_ETSHFrame.scroll
+set td_ETFillH		$td_ETSHFrame.fill
 set td_ETClose		$td_ET.close
 
 # }}}
@@ -1458,7 +1509,7 @@ set td_ETClose		$td_ET.close
 
 if [winfo exists $td_Top] {destroy $td_Top}
 
-toplevel $td_Top -class TDebug -borderwidth 2 
+toplevel $td_Top -class TDebug -borderwidth 0
 wm title $td_Top "TDebug for [winfo name .]"
 wm withdraw $td_Top
 
@@ -1535,18 +1586,19 @@ tk_menuBar $td_Menubar $td_MBDebug $td_MBOptions $td_MBSelection \
 tk_bindForTraversal $td_Top
 
 #We need the input focus when the cursor is inside!
-if {"[focus default]" == "none"} {focus default .}
-bind $td_Top <FocusIn> {
-    if {"%d" == "NotifyVirtual"} {
-	set td_priv(focus) [focus default]
-	focus default %W
-	focus %W
-}   }
-bind $td_Top <FocusOut> {
-    if {"%d" == "NotifyVirtual"} {
-	focus default $td_priv(focus)
-	focus [focus default]
-}   }
+if {$tk_version < 4.0} {
+    if {"[focus default]" == "none"} {focus default .}
+    bind $td_Top <FocusIn> {
+	if {"%d" == "NotifyVirtual"} {
+	    set td_priv(focus) [focus default]
+	    focus default %W
+	    focus %W
+    }   }
+    bind $td_Top <FocusOut> {
+	if {"%d" == "NotifyVirtual"} {
+	    focus default $td_priv(focus)
+	    focus [focus default]
+}   }   }
 
 # bind Accelerators
 bind $td_Top <Control-b> {wm deiconify $td_BT}
@@ -1582,24 +1634,52 @@ frame $td_ListNameFrame -relief raised -borderwidth 0
 pack $td_ListNameFrame -side top -fill x
 label $td_ListName -text "Proc  :" -borderwidth 0 -width 8
 pack $td_ListName -side left
-entry $td_ListProc -textvariable td_priv(proc) -relief groove -state disabled
+if {$tk_version >= 4.0} {
+    entry $td_ListProc -textvariable td_priv(proc) -relief groove \
+	    -state disabled -highlightthickness 0
+} else {
+    entry $td_ListProc -textvariable td_priv(proc) -relief groove -state disabled
+}
 pack $td_ListProc -side left -expand 1 -fill x
 
 frame $td_TextFrame -relief raised -borderwidth 0
 pack $td_TextFrame -side top -expand 1 -fill both
 
-if {$td_priv(constrainscroll)} {
-    scrollbar $td_ListScroll -command "td_constrainScroll"
+if {$tk_version < 4.0 && $td_priv(constrainscroll)} {
+    scrollbar $td_ListScroll -command "td_constrainScroll" -relief sunken
 } else {
-    scrollbar $td_ListScroll -command "$td_Listing yview"
+    scrollbar $td_ListScroll -command "$td_Listing yview" -relief sunken
 }
 pack $td_ListScroll -side $td_priv(scrollbarside) -fill y
 
-text $td_Listing -width 27 -height 2 \
-	-relief sunken -setgrid 1 -wrap none -state disabled \
-	-borderwidth 2  -yscrollcommand td_catchScroll
+if {$tk_version >= 4.0} {
+    text $td_Listing -width 27 -height 2 \
+	    -relief sunken -setgrid 1 -wrap none -state disabled \
+	    -borderwidth 2  -yscrollcommand "$td_ListScroll set"
+} else {
+    text $td_Listing -width 27 -height 2 \
+	    -relief sunken -setgrid 1 -wrap none -state disabled \
+	    -borderwidth 2  -yscrollcommand td_catchScroll
+}
 pack $td_Listing -side $td_priv(scrollbarside) -expand 1 -fill both
 
+if {$tk_version >= 4.0} {
+    frame $td_ListHSFrame -borderwidth 0
+    pack $td_ListHSFrame -side top -fill x
+
+    scrollbar $td_ListHScroll -orient horizontal -relief sunken \
+	    -command "$td_Listing xview"
+    $td_Listing configure -xscrollcommand "$td_ListHScroll set"
+
+    frame $td_ListHFill \
+	    -width [expr [$td_ListScroll cget -width] + 4 + \
+	                 2 * [$td_ListScroll cget -highlightthickness]] \
+	    -height [expr [$td_ListHScroll cget -width] + 4 + \
+	                  2 * [$td_ListHScroll cget -highlightthickness]]
+    pack $td_ListHFill -side $td_priv(scrollbarside)
+    pack $td_ListHScroll -side $td_priv(scrollbarside) -expand 1 -fill x
+}    
+
 eval $td_Listing tag configure prepared $td_priv(preparedtag)
 eval $td_Listing tag configure active $td_priv(activetag)
 eval $td_Listing tag configure break $td_priv(breaktag)
@@ -1617,7 +1697,7 @@ pack $td_VarFrame -side left -fill both
 frame $td_VarFrame1 -borderwidth 2 -relief raised
 pack $td_VarFrame1 -side top -fill both -expand 1 -padx 2 -pady 2
 
-label $td_VarName -text "Variables: " -borderwidth 0
+label $td_VarName -text "Variables: " -borderwidth 2
 pack $td_VarName -side top -fill x
 
 frame $td_VarFrame2 
@@ -1626,12 +1706,17 @@ pack $td_VarFrame2 -side top -expand 1 -fill both
 frame $td_VarFrame3
 pack $td_VarFrame3 -side top -expand 1 -fill both
 
-scrollbar $td_VarScrollY -command "$td_Vars yview" -relief flat
+scrollbar $td_VarScrollY -command "$td_Vars yview" -relief sunken
 pack $td_VarScrollY -side $td_priv(scrollbarside) -fill y
 
-listbox $td_Vars -xscrollcommand "$td_VarScrollX set" \
-	-relief sunken -geometry $td_priv(varwidth)x2
-if {$td_priv(constrainscroll)} {
+if {$tk_version >= 4.0} {
+    listbox $td_Vars -xscrollcommand "$td_VarScrollX set" \
+	    -relief sunken -width $td_priv(varwidth) -height 2
+} else {
+    listbox $td_Vars -xscrollcommand "$td_VarScrollX set" \
+	    -relief sunken -geometry $td_priv(varwidth)x2
+}
+if {$tk_version < 4 && $td_priv(constrainscroll)} {
     $td_Vars configure -yscrollcommand td_catchVarScroll
 } else {
     $td_Vars configure -yscrollcommand "$td_VarScrollY set" 
@@ -1641,9 +1726,20 @@ pack $td_Vars -side $td_priv(scrollbarside) -expand 1 -fill both
 frame $td_VarFrame4
 pack $td_VarFrame4 -side top -fill x
 
-scrollbar $td_VarScrollX -command "$td_Vars xview" -orient horiz -relief flat
-frame $td_VarFrame5 -width [lindex [$td_VarScrollY configure -width] 4] \
-	-height [lindex [$td_VarScrollX configure -width] 4]
+scrollbar $td_VarScrollX -command "$td_Vars xview" -orient horiz -relief sunken
+
+if {$tk_version >= 4.0} {
+    frame $td_VarFrame5 \
+	    -width [expr [$td_VarScrollY cget -width] + 4 + \
+	                 2 * [$td_VarScrollY cget -highlightthickness]] \
+	    -height [expr [$td_VarScrollX cget -width] + 4 + \
+	                  2 * [$td_VarScrollX cget -highlightthickness]]
+} else {
+    frame $td_VarFrame5 \
+	    -width [expr [lindex [$td_VarScrollY configure -width] 4] + 4] \
+	    -height [expr [lindex [$td_VarScrollX configure -width] 4] + 4]
+}
+
 pack $td_VarFrame5 -side $td_priv(scrollbarside)
 pack $td_VarScrollX -side $td_priv(scrollbarside) -expand 1 -fill x
 
@@ -1658,7 +1754,11 @@ pack $td_ResultFrame -side top -fill x
 label $td_ResultName -relief flat -text Result: -borderwidth 0 -width 8
 pack $td_ResultName -side left
 
-entry $td_Result -relief groove -state disabled
+if {$tk_version >= 4.0} {
+    entry $td_Result -relief groove -state disabled -highlightthickness 0
+} else {
+    entry $td_Result -relief groove -state disabled
+}
 pack $td_Result -side left -expand 1 -fill x
 
 # }}}
@@ -1681,32 +1781,61 @@ bind $td_Eval <Control-c> {wm positionfrom $td_Top user ; wm withdraw $td_Top}
 # {{{ the buttons
 
 frame $td_Buttons
-pack $td_Buttons -side top -fill x -padx 2 -pady 2
+radiobutton $td_BStop -text Stop -width 4 -variable td_priv(state) -value stop \
+	-relief raised
 
-radiobutton $td_BStop -text Stop -width 4 -variable td_priv(state) -value stop
-pack $td_BStop -side left -expand 1 -fill x
-radiobutton $td_BNext -text Next -width 4 -variable td_priv(state) -value next
-pack $td_BNext -side left -expand 1 -fill x
-radiobutton $td_BSlow -text Slow -width 4 -variable td_priv(state) -value slow
-pack $td_BSlow -side left -expand 1 -fill x
-radiobutton $td_BFast -text Fast -width 4 -variable td_priv(state) -value fast
-pack $td_BFast -side left -expand 1 -fill x
-radiobutton $td_BNonstop -text Nonstop -width 4 -variable td_priv(state) -value nonstop
-pack $td_BNonstop -side left -expand 1 -fill x
-radiobutton $td_BBreak -text Break -width 4 -variable td_priv(state) -value break
-pack $td_BBreak -side left -expand 1 -fill x
+radiobutton $td_BNext -text Next -width 4 -variable td_priv(state) -value next \
+	-relief raised
+radiobutton $td_BSlow -text Slow -width 4 -variable td_priv(state) -value slow \
+	-relief raised
+radiobutton $td_BFast -text Fast -width 4 -variable td_priv(state) -value fast \
+	-relief raised
+radiobutton $td_BNonstop -text Nonstop -width 4 -variable td_priv(state) \
+	-value nonstop -relief raised
+radiobutton $td_BBreak -text Break -width 4 -variable td_priv(state) -value break \
+	-relief raised
+
+if {$tk_version >= 4.0} {
+    pack $td_Buttons -side top -fill x
+    pack $td_BStop -side left -expand 1 -fill x
+    pack $td_BNext -side left -expand 1 -fill x
+    pack $td_BSlow -side left -expand 1 -fill x
+    pack $td_BFast -side left -expand 1 -fill x
+    pack $td_BNonstop -side left -expand 1 -fill x
+    pack $td_BBreak -side left -expand 1 -fill x
+} else {
+    pack $td_Buttons -side top -fill x -pady 2
+    pack $td_BStop -side left -expand 1 -fill x -padx 2
+    pack $td_BNext -side left -expand 1 -fill x -padx 2
+    pack $td_BSlow -side left -expand 1 -fill x -padx 2
+    pack $td_BFast -side left -expand 1 -fill x -padx 2
+    pack $td_BNonstop -side left -expand 1 -fill x -padx 2
+    pack $td_BBreak -side left -expand 1 -fill x -padx 2
+}
 
 # }}}
 # {{{ the delay
 
-frame $td_DelayFrame
-pack $td_DelayFrame -side top -fill x -padx 2 -pady 2
+# the following layout is different depending on the button's highlightframes
+if {$tk_version >= 4.0} {
+    frame $td_DelayFrame
+    pack $td_DelayFrame -side top -fill x -padx 0 -pady 0
 
-frame $td_DelayFrame1 -borderwidth 2 -relief raised
-pack $td_DelayFrame1 -side left -expand 1 -fill x
+    frame $td_DelayFrame1 -borderwidth 2 -relief raised
+    pack $td_DelayFrame1 -side left -expand 1 -fill x -padx 2 -pady 2
 
-label $td_Delay -text "Delay:  $td_priv(delay)" -width 11 -borderwidth 1
-pack $td_Delay -side left -expand 1 -fill both
+    label $td_Delay -text "Delay:  $td_priv(delay)" -width 11 -borderwidth 4
+    pack $td_Delay -side left -expand 1 -fill both
+} else {
+    frame $td_DelayFrame
+    pack $td_DelayFrame -side top -fill x -pady 2
+
+    frame $td_DelayFrame1 -borderwidth 2 -relief raised
+    pack $td_DelayFrame1 -side left -expand 1 -fill x -padx 2
+
+    label $td_Delay -text "Delay:  $td_priv(delay)" -width 11 -borderwidth 1
+    pack $td_Delay -side left -expand 1 -fill both
+}
 
 button $td_DelayLess -text "-" -width 2 -relief raised -command {
     global td_priv
@@ -1715,7 +1844,6 @@ button $td_DelayLess -text "-" -width 2 -relief raised -command {
 	$td_Delay configure -text "Delay: [format %4d $td_priv(delay)]"
     }
 }
-pack $td_DelayLess -side left -fill both
 
 button $td_DelayMore -text "+" -width 2 -relief raised -command {
     global td_priv
@@ -1724,7 +1852,13 @@ button $td_DelayMore -text "+" -width 2 -relief raised -command {
 	$td_Delay configure -text "Delay: [format %4d $td_priv(delay)]"
     }
 }
-pack $td_DelayMore -side left -fill both
+if {$tk_version >= 4.0} {
+    pack $td_DelayLess -side left -fill both
+    pack $td_DelayMore -side left -fill both
+} else {
+    pack $td_DelayLess -side left -fill both -padx 2
+    pack $td_DelayMore -side left -fill both -padx 2
+}
 
 # }}}
 
@@ -1737,20 +1871,48 @@ wm minsize $td_Top 27 2
 
 # {{{ The Backtrace toplevel
 
-toplevel $td_BT -borderwidth 2
+toplevel $td_BT
 wm withdraw $td_BT
 wm title $td_BT TDebug-Backtrace
+
 frame $td_BTMain -relief raised -borderwidth 2
 pack $td_BTMain -expand 1 -fill both -padx 2 -pady 2
-scrollbar $td_BTScroll -command "$td_BTText yview" -relief flat
+
+frame $td_BTTop
+pack $td_BTTop -side top -expand 1 -fill both
+
+scrollbar $td_BTScroll -command "$td_BTText yview" -relief sunken
 pack $td_BTScroll -side $td_priv(scrollbarside) -fill y
+
 text $td_BTText -relief sunken -borderwidth 2 -width 30 -height 3 -setgrid 1 \
 	-wrap none -state disabled
 pack $td_BTText -side $td_priv(scrollbarside) -expand 1 -fill both
 $td_BTText configure -yscrollcommand "$td_BTScroll set"
+
+if {$tk_version >= 4.0} {
+    frame $td_BTSHFrame -borderwidth 0
+    pack $td_BTSHFrame -side top -fill x
+
+    scrollbar $td_BTScrollH -orient horizontal -relief sunken \
+	    -command "$td_BTText xview"
+    $td_BTText configure -xscrollcommand "$td_BTScrollH set"
+
+    frame $td_BTFillH \
+	    -width [expr [$td_BTScroll cget -width] + 4 + \
+	                 2 * [$td_BTScroll cget -highlightthickness]] \
+	    -height [expr [$td_BTScrollH cget -width] + 4 + \
+	                  2 * [$td_BTScrollH cget -highlightthickness]]
+    pack $td_BTFillH -side $td_priv(scrollbarside)
+    pack $td_BTScrollH -side $td_priv(scrollbarside) -expand 1 -fill x
+}
+
 button $td_BTClose -text Close -relief raised \
 	-command "wm positionfrom $td_BT user ;	wm withdraw $td_BT"
-pack $td_BTClose -side top -fill x -padx 2 -pady 2
+if {$tk_version >= 4.0} {
+    pack $td_BTClose -side top -fill x
+} else {
+    pack $td_BTClose -side top -fill x -padx 2 -pady 2
+}
 
 wm geometry $td_BT $td_priv(backtracewidth)x$td_priv(backtraceheight)
 if [info exists td_priv(backtracegeometry)] {
@@ -1760,33 +1922,63 @@ if [info exists td_priv(backtracegeometry)] {
 
 bind $td_BT <Control-c> {wm positionfrom $td_BT user ; wm withdraw $td_BT}
 bind $td_BT <Control-w> {$td_MenuOptions invoke 3}
+
 #We need the input focus when the cursor is inside!
-if {"[focus default]" == "none"} {focus default .}
-bind $td_BT <FocusIn> {
-    if {"%d" == "NotifyVirtual"} {focus %W}
-}
-bind $td_BT <FocusOut> {
-    if {"%d" == "NotifyVirtual"} {focus [focus default]}
+if {$tk_version < 4.0} {
+    if {"[focus default]" == "none"} {focus default .}
+    bind $td_BT <FocusIn> {
+	if {"%d" == "NotifyVirtual"} {focus %W}
+    }
+    bind $td_BT <FocusOut> {
+	if {"%d" == "NotifyVirtual"} {focus [focus default]}
+    }
 }
 
 # }}}
 # {{{ The Widget Hierarchy toplevel
 
-toplevel $td_WH -borderwidth 2
+toplevel $td_WH
 wm withdraw $td_WH
 wm title $td_WH TDebug-Widget-Hierarchy
+
 frame $td_WHMain -relief raised -borderwidth 2
 pack $td_WHMain -expand 1 -fill both -padx 2 -pady 2
-scrollbar $td_WHScroll -command "$td_WHText yview" -relief flat
+
+frame $td_WHTop
+pack $td_WHTop -side top -expand 1 -fill both
+
+scrollbar $td_WHScroll -command "$td_WHText yview" -relief sunken
 pack $td_WHScroll -side $td_priv(scrollbarside) -fill y
+
 text $td_WHText -relief sunken -borderwidth 2 -width 30 -height 3 -setgrid 1 \
 	-wrap none -state disabled 
 pack $td_WHText -side $td_priv(scrollbarside) -expand 1 -fill both
 $td_WHText configure -yscrollcommand "$td_WHScroll set"
+
+if {$tk_version >= 4.0} {
+    frame $td_WHSHFrame -borderwidth 0
+    pack $td_WHSHFrame -side top -fill x
+
+    scrollbar $td_WHScrollH -orient horizontal -relief sunken \
+	    -command "$td_WHText xview"
+    $td_WHText configure -xscrollcommand "$td_WHScrollH set"
+
+    frame $td_WHFillH \
+	    -width [expr [$td_WHScroll cget -width] + 4 + \
+	                 2 * [$td_WHScroll cget -highlightthickness]] \
+	    -height [expr [$td_WHScrollH cget -width] + 4 + \
+	                  2 * [$td_WHScrollH cget -highlightthickness]]
+    pack $td_WHFillH -side $td_priv(scrollbarside)
+    pack $td_WHScrollH -side $td_priv(scrollbarside) -expand 1 -fill x
+}
+
 button $td_WHClose -text Close -relief raised \
 	-command "wm positionfrom $td_WH user ;	wm withdraw $td_WH"
-pack $td_WHClose -side top -fill x -padx 2 -pady 2
-
+if {$tk_version >= 4.0} {
+    pack $td_WHClose -side top -fill x
+} else {
+    pack $td_WHClose -side top -fill x -padx 2 -pady 2
+}
 wm geometry $td_WH $td_priv(widgetswidth)x$td_priv(widgetsheight)
 if [info exists td_priv(widgetsgeometry)] {
     wm geometry $td_WH $td_priv(widgetsgeometry)
@@ -1796,31 +1988,61 @@ if [info exists td_priv(widgetsgeometry)] {
 bind $td_WH <Control-c> {wm positionfrom $td_WH user ; wm withdraw $td_WH}
 bind $td_WH <Control-f> {$td_MenuOptions invoke 4}
 #We need the input focus when the cursor is inside!
-if {"[focus default]" == "none"} {focus default .}
-bind $td_WH <FocusIn> {
-    if {"%d" == "NotifyVirtual"} {focus %W}
-}
-bind $td_WH <FocusOut> {
-    if {"%d" == "NotifyVirtual"} {focus [focus default]}
+if {$tk_version < 4.0} {
+    if {"[focus default]" == "none"} {focus default .}
+    bind $td_WH <FocusIn> {
+	if {"%d" == "NotifyVirtual"} {focus %W}
+    }
+    bind $td_WH <FocusOut> {
+	if {"%d" == "NotifyVirtual"} {focus [focus default]}
+    }
 }
 
 # }}}
 # {{{ The Error Trace toplevel
 
-toplevel $td_ET -borderwidth 2
+toplevel $td_ET
 wm withdraw $td_ET
 wm title $td_ET TDebug-ErrorTrace
+
 frame $td_ETMain -relief raised -borderwidth 2
 pack $td_ETMain -expand 1 -fill both -padx 2 -pady 2
-scrollbar $td_ETScroll -command "$td_ETText yview" -relief flat
+
+frame $td_ETTop
+pack $td_ETTop -side top -expand 1 -fill both
+
+scrollbar $td_ETScroll -command "$td_ETText yview" -relief sunken
 pack $td_ETScroll -side $td_priv(scrollbarside) -fill y
+
 text $td_ETText -relief sunken -borderwidth 2 -width 30 -height 3 -setgrid 1 \
 	-wrap none -state disabled
 pack $td_ETText -side $td_priv(scrollbarside) -expand 1 -fill both
 $td_ETText configure -yscrollcommand "$td_ETScroll set"
+
+if {$tk_version >= 4.0} {
+    frame $td_ETSHFrame -borderwidth 0
+    pack $td_ETSHFrame -side top -fill x
+
+    scrollbar $td_ETScrollH -orient horizontal -relief sunken \
+	    -command "$td_ETText xview"
+    $td_ETText configure -xscrollcommand "$td_ETScrollH set"
+
+    frame $td_ETFillH \
+	    -width [expr [$td_ETScroll cget -width] + 4 + \
+	                 2 * [$td_ETScroll cget -highlightthickness]] \
+	    -height [expr [$td_ETScrollH cget -width] + 4 + \
+	                  2 * [$td_ETScrollH cget -highlightthickness]]
+    pack $td_ETFillH -side $td_priv(scrollbarside)
+    pack $td_ETScrollH -side $td_priv(scrollbarside) -expand 1 -fill x
+}
+
 button $td_ETClose -text Close -relief raised \
 	-command "wm positionfrom $td_ET user ;	wm withdraw $td_ET"
-pack $td_ETClose -side top -fill x -padx 2 -pady 2
+if {$tk_version >= 4.0} {
+    pack $td_ETClose -side top -fill x
+} else {
+    pack $td_ETClose -side top -fill x -padx 2 -pady 2
+}
 
 wm geometry $td_ET $td_priv(errorwidth)x$td_priv(errorheight)
 if [info exists td_priv(errorgeometry)] {
@@ -1830,17 +2052,17 @@ if [info exists td_priv(errorgeometry)] {
 
 bind $td_ET <Control-c> {wm positionfrom $td_ET user ; wm withdraw $td_ET}
 #We need the input focus when the cursor is inside!
-if {"[focus default]" == "none"} {focus default .}
-bind $td_ET <FocusIn> {
-    if {"%d" == "NotifyVirtual"} {focus %W}
-}
-bind $td_ET <FocusOut> {
-    if {"%d" == "NotifyVirtual"} {focus [focus default]}
-}
+if {$tk_version < 4.0} {
+    if {"[focus default]" == "none"} {focus default .}
+    bind $td_ET <FocusIn> {
+	if {"%d" == "NotifyVirtual"} {focus %W}
+    }
+    bind $td_ET <FocusOut> {
+	if {"%d" == "NotifyVirtual"} {focus [focus default]}
+}   }
 
 # }}}
 
-
 # }}}
 
 # install new `proc'
